import numpy as np


def generate_masks_(m: int, n: int, s: float) -> np.ndarray:
    """Generates set of binary masks with properties defined by n, m, s params.

    Results of this function are stochastic, that is, calls with the same sets
    of arguments might generate outputs of different shapes. Check generate_masks
    and generation_wrapper function for more deterministic behaviour.

    :param m: int, number of ones in each mask
    :param n: int, number of masks in the set
    :param s: float, scale param controls overlap of generated masks
    :return: np.ndarray, matrix of binary vectors
    """

    total_positions = int(m * s)
    masks = []

    for _ in range(n):
        new_vector = np.zeros([total_positions])
        idx = np.random.choice(range(total_positions), m, replace=False)
        new_vector[idx] = 1
        masks.append(new_vector)

    masks = np.array(masks)
    # drop useless positions
    masks = masks[:, ~np.all(masks == 0, axis=0)]
    return masks


def generate_masks(m: int, n: int, s: float) -> np.ndarray:
    """Generates set of binary masks with properties defined by n, m, s params.

    Resulting masks are required to have fixed features size as it's described in [1].
    Since process of masks generation is stochastic therefore function evaluates
    generate_masks_ multiple times till expected size is acquired.

    :param m: int, number of ones in each mask
    :param n: int, number of masks in the set
    :param s: float, scale param controls overlap of generated masks
    :return: np.ndarray, matrix of binary vectors

    References

    [1] `Masksembles for Uncertainty Estimation: Supplementary Material`,
    Nikita Durasov, Timur Bagautdinov, Pierre Baque, Pascal Fua
    """

    masks = generate_masks_(m, n, s)
    # hardcoded formula for expected size, check reference
    expected_size = int(m * s * (1 - (1 - 1 / s) ** n))
    while masks.shape[1] != expected_size:
        masks = generate_masks_(m, n, s)
    return masks


def generation_wrapper(c: int, n: int, scale: float) -> np.ndarray:
    """Generates set of binary masks with properties defined by c, n, scale params.

     Allows to generate masks sets with predefined features number c. Particularly
     convenient to use in torch-like layers where one need to define shapes inputs
     tensors beforehand.

    :param c: int, number of channels in generated masks
    :param n: int, number of masks in the set
    :param scale: float, scale param controls overlap of generated masks
    :return: np.ndarray, matrix of binary vectors
    """

    if c < 10:
        raise ValueError("Masksembles approach couldn't be used in such setups where "
                         f"number of channels is less then 10. Current value is (channels={c}). "
                         "Please increase number of features in your layer or remove this "
                         "particular instance of Masksembles from your architecture.")

    if scale > 6.:
        raise ValueError("Masksembles approach couldn't be used in such setups where "
                         f"scale parameter is larger then 6. Current value is (scale={scale}).")
    
    # inverse formula for number of active features in masks
    active_features = int(int(c) / (scale * (1 - (1 - 1 / scale) ** n)))

    # Fix the last part by using binary search 
    max_iter = 1000

    min = np.max([scale * 0.8, 1.0])
    max = scale * 1.2

    for _ in range(max_iter):
        mid = (min + max) / 2
        masks = generate_masks(active_features, n, mid)
        if masks.shape[-1] == c:
            break
        elif masks.shape[-1] > c:
            max = mid
        else:
            min = mid

    if masks.shape[-1] != c:
        raise ValueError("generation_wrapper function failed to generate masks with "
                         "requested number of features. Please try to change scale parameter")

    return masks



import unittest
import tensorflow as tf

import masksembles.keras


class TestCreation(unittest.TestCase):

    def test_init_failed(self):
        layer = masksembles.keras.Masksembles2D(4, 11.)
        self.assertRaises(ValueError, layer, tf.ones([4, 10, 4, 4]))

    def test_init_success(self):
        layer = masksembles.keras.Masksembles2D(4, 11.)



import torch
from torch import nn

from . import common


class Masksembles2D(nn.Module):
    """
    :class:`Masksembles2D` is high-level class that implements Masksembles approach
    for 2-dimensional inputs (similar to :class:`torch.nn.Dropout2d`).

    :param channels: int, number of channels used in masks.
    :param n: int, number of masks
    :param scale: float, scale parameter similar to *S* in [1]. Larger values decrease \
        subnetworks correlations but at the same time decrease capacity of every individual model.

    Shape:
        * Input: (N, C, H, W)
        * Output: (N, C, H, W) (same shape as input)

    Examples:

    >>> m = Masksembles2D(16, 4, 2.0)
    >>> input = torch.ones([4, 16, 28, 28])
    >>> output = m(input)

    References:

    [1] `Masksembles for Uncertainty Estimation`,
    Nikita Durasov, Timur Bagautdinov, Pierre Baque, Pascal Fua

    """
    def __init__(self, channels: int, n: int, scale: float):
        super().__init__()
        self.channels, self.n, self.scale = channels, n, scale
        masks_np = common.generation_wrapper(channels, n, scale)  # numpy float64 by default
        masks = torch.as_tensor(masks_np, dtype=torch.float32)    # make float32 here
        self.register_buffer('masks', masks, persistent=False)    # not trainable, moves with .to()

    def forward(self, inputs):
        # make sure masks match dtype/device (usually already true because of buffer, but safe)
        masks = self.masks.to(dtype=inputs.dtype, device=inputs.device)

        batch = inputs.shape[0]
        # safer split even if batch % n != 0
        chunks = torch.chunk(inputs.unsqueeze(1), self.n, dim=0)  # returns nearly equal chunks
        x = torch.cat(chunks, dim=1).permute(1, 0, 2, 3, 4)       # [n, ?, C, H, W]
        x = x * masks.unsqueeze(1).unsqueeze(-1).unsqueeze(-1)    # broadcast masks
        x = torch.cat(torch.split(x, 1, dim=0), dim=1)
        return x.squeeze(0)



class Masksembles1D(nn.Module):
    """
    :class:`Masksembles1D` is high-level class that implements Masksembles approach
    for 1-dimensional inputs (similar to :class:`torch.nn.Dropout`).

    :param channels: int, number of channels used in masks.
    :param n: int, number of masks
    :param scale: float, scale parameter similar to *S* in [1]. Larger values decrease \
        subnetworks correlations but at the same time decrease capacity of every individual model.

    Shape:
        * Input: (N, C)
        * Output: (N, C) (same shape as input)

    Examples:

    >>> m = Masksembles1D(16, 4, 2.0)
    >>> input = torch.ones([4, 16])
    >>> output = m(input)


    References:

    [1] `Masksembles for Uncertainty Estimation`,
    Nikita Durasov, Timur Bagautdinov, Pierre Baque, Pascal Fua

    """

    def __init__(self, channels: int, n: int, scale: float):
        super().__init__()
        self.channels, self.n, self.scale = channels, n, scale
        masks_np = common.generation_wrapper(channels, n, scale)
        masks = torch.as_tensor(masks_np, dtype=torch.float32)
        self.register_buffer('masks', masks, persistent=False)

    def forward(self, inputs):
        masks = self.masks.to(dtype=inputs.dtype, device=inputs.device)

        batch = inputs.shape[0]
        chunks = torch.chunk(inputs.unsqueeze(1), self.n, dim=0)
        x = torch.cat(chunks, dim=1).permute(1, 0, 2)             # [n, ?, C]
        x = x * masks.unsqueeze(1)
        x = torch.cat(torch.split(x, 1, dim=0), dim=1)
        return x.squeeze(0)



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
readme = (HERE / "README.md").read_text(encoding="utf-8")

setup(
    name="masksembles",                 # change if taken on PyPI
    version="1.1.1",                    # use semantic x.y.z (PyPI won’t accept re-uploads)
    author="Nikita Durasov",
    author_email="yassnda@gmail.com",
    description="Official implementation of Masksembles approach",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/nikitadurasov/masksembles",
    project_urls={
        "Issues": "https://github.com/nikitadurasov/masksembles/issues",
        "Documentation": "https://github.com/nikitadurasov/masksembles#readme",
    },
    license="MIT",
    packages=find_packages(exclude=("tests", "test", "notebooks")),
    python_requires=">=3.8",
    install_requires=[
        "numpy>=1.20",
    ],
    extras_require={
        "torch": ["torch>=1.12"],
        "tensorflow": ["tensorflow>=2.9"],
        "dev": ["pytest", "black", "flake8"],
    },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=["uncertainty", "ensembles", "deep-learning", "pytorch", "tensorflow"],
)



import tensorflow as tf
from . import common


class Masksembles2D(tf.keras.layers.Layer):
    """
    :class:`Masksembles2D` is high-level class that implements Masksembles approach
    for 2-dimensional inputs (similar to :class:`tensorflow.keras.layers.SpatialDropout1D`).

    :param n: int, number of masks
    :param scale: float, scale parameter similar to *S* in [1]. Larger values decrease \
        subnetworks correlations but at the same time decrease capacity of every individual model.

    Shape:
        * Input: (N, H, W, C)
        * Output: (N, H, W, C) (same shape as input)

    Examples:

    >>> m = Masksembles2D(4, 2.0)
    >>> inputs = tf.ones([4, 28, 28, 16])
    >>> output = m(inputs)

    References:

    [1] `Masksembles for Uncertainty Estimation`,
    Nikita Durasov, Timur Bagautdinov, Pierre Baque, Pascal Fua

    """

    def __init__(self, n: int, scale: float):
        super(Masksembles2D, self).__init__()

        self.n = n
        self.scale = scale

    def build(self, input_shape):
        channels = input_shape[-1]
        masks = common.generation_wrapper(channels, self.n, self.scale)
        self.masks = self.add_weight("masks",
                                     shape=masks.shape,
                                     trainable=False,
                                     dtype="float32")
        self.masks.assign(masks)

    def call(self, inputs, training=False):
        # inputs : [N, H, W, C]
        # masks : [M, C]
        x = tf.stack(tf.split(inputs, self.n))
        # x : [M, N // M, H, W, C]
        # masks : [M, 1, 1, 1, C]
        x = x * self.masks[:, tf.newaxis, tf.newaxis, tf.newaxis]
        x = tf.concat(tf.split(x, self.n), axis=1)
        return tf.squeeze(x, axis=0)


class Masksembles1D(tf.keras.layers.Layer):
    """
    :class:`Masksembles1D` is high-level class that implements Masksembles approach
    for 1-dimensional inputs (similar to :class:`tensorflow.keras.layers.Dropout`).

    :param n: int, number of masks
    :param scale: float, scale parameter similar to *S* in [1]. Larger values decrease \
        subnetworks correlations but at the same time decrease capacity of every individual model.

    Shape:
        * Input: (N, C)
        * Output: (N, C) (same shape as input)

    Examples:

    >>> m = Masksembles1D(4, 2.0)
    >>> inputs = tf.ones([4, 16])
    >>> output = m(inputs)


    References:

    [1] `Masksembles for Uncertainty Estimation`,
    Nikita Durasov, Timur Bagautdinov, Pierre Baque, Pascal Fua

    """

    def __init__(self, n: int, scale: float):
        super(Masksembles1D, self).__init__()

        self.n = n
        self.scale = scale

    def build(self, input_shape):
        channels = input_shape[-1]
        masks = common.generation_wrapper(channels, self.n, self.scale)
        self.masks = self.add_weight("masks",
                                     shape=masks.shape,
                                     trainable=False,
                                     dtype="float32")
        self.masks.assign(masks)

    def call(self, inputs, training=False):
        x = tf.stack(tf.split(inputs, self.n))
        x = x * self.masks[:, tf.newaxis]
        x = tf.concat(tf.split(x, self.n), axis=1)
        return tf.squeeze(x, axis=0)
